"""
training code
"""
from __future__ import absolute_import
from __future__ import division
import argparse
import os
import torch

from config import cfg, assert_and_infer_cfg
import loss
import network
import optimizer
import time
import numpy as np
import random
import torchvision.transforms as transforms
from collections import namedtuple

class_colors = [  # [  0,   0,   0],
    [128, 64, 128],
    [244, 35, 232],
    [70, 70, 70],
    [102, 102, 156],
    [190, 153, 153],
    [153, 153, 153],
    [250, 170, 30],
    [220, 220, 0],
    [107, 142, 35],
    [152, 251, 152],
    [70, 130, 180],
    [220, 20, 60],
    [255, 0, 0],
    [0, 0, 142],
    [0, 0, 70],
    [0, 60, 100],
    [0, 80, 100],
    [0, 0, 230],
    [119, 11, 32],
]

# Argument Parser
parser = argparse.ArgumentParser(description='Semantic Segmentation')
parser.add_argument('--lr', type=float, default=0.01)
parser.add_argument('--arch', type=str, default='network.deepv3.DeepR50V3PlusD',
                    help='Network architecture. We have DeepSRNX50V3PlusD (backbone: ResNeXt50) \
                    and deepWV3Plus (backbone: WideResNet38).')
parser.add_argument('--dataset', nargs='*', type=str, default=['cityscapes'],
                    help='a list of datasets; cityscapes, mapillary, camvid, kitti, gtav, mapillary, synthia')
parser.add_argument('--image_uniform_sampling', action='store_true', default=False,
                    help='uniformly sample images across the multiple source domains')
parser.add_argument('--val_dataset', nargs='*', type=str, default=['bdd100k'],
                    help='a list consists of cityscapes, mapillary, gtav, bdd100k, synthia')
parser.add_argument('--covstat_val_dataset', nargs='*', type=str, default=['cityscapes'],
                    help='a list consists of cityscapes, mapillary, gtav, bdd100k, synthia')
parser.add_argument('--cv', type=int, default=0,
                    help='cross-validation split id to use. Default # of splits set to 3 in config')
parser.add_argument('--class_uniform_pct', type=float, default=0,
                    help='What fraction of images is uniformly sampled')
parser.add_argument('--class_uniform_tile', type=int, default=1024,
                    help='tile size for class uniform sampling')
parser.add_argument('--coarse_boost_classes', type=str, default=None,
                    help='use coarse annotations to boost fine data with specific classes')

parser.add_argument('--img_wt_loss', action='store_true', default=False,
                    help='per-image class-weighted loss')
parser.add_argument('--cls_wt_loss', action='store_true', default=False,
                    help='class-weighted loss')
parser.add_argument('--batch_weighting', action='store_true', default=False,
                    help='Batch weighting for class (use nll class weighting using batch stats')

parser.add_argument('--jointwtborder', action='store_true', default=False,
                    help='Enable boundary label relaxation')
parser.add_argument('--strict_bdr_cls', type=str, default='',
                    help='Enable boundary label relaxation for specific classes')
parser.add_argument('--rlx_off_iter', type=int, default=-1,
                    help='Turn off border relaxation after specific epoch count')
parser.add_argument('--rescale', type=float, default=1.0,
                    help='Warm Restarts new learning rate ratio compared to original lr')
parser.add_argument('--repoly', type=float, default=1.5,
                    help='Warm Restart new poly exp')

parser.add_argument('--fp16', action='store_true', default=False,
                    help='Use Nvidia Apex AMP')
parser.add_argument('--local_rank', default=0, type=int,
                    help='parameter used by apex library')

parser.add_argument('--sgd', action='store_true', default=True)
parser.add_argument('--adam', action='store_true', default=False)
parser.add_argument('--amsgrad', action='store_true', default=False)

parser.add_argument('--freeze_trunk', action='store_true', default=False)
parser.add_argument('--hardnm', default=0, type=int,
                    help='0 means no aug, 1 means hard negative mining iter 1,' +
                    '2 means hard negative mining iter 2')

parser.add_argument('--trunk', type=str, default='resnet101',
                    help='trunk model, can be: resnet101 (default), resnet50')
parser.add_argument('--max_epoch', type=int, default=180)
parser.add_argument('--max_iter', type=int, default=30000)
parser.add_argument('--max_cu_epoch', type=int, default=100000,
                    help='Class Uniform Max Epochs')
parser.add_argument('--start_epoch', type=int, default=0)
parser.add_argument('--crop_nopad', action='store_true', default=False)
parser.add_argument('--rrotate', type=int,
                    default=0, help='degree of random roate')
parser.add_argument('--color_aug', type=float,
                    default=0.0, help='level of color augmentation')
parser.add_argument('--gblur', action='store_true', default=False,
                    help='Use Guassian Blur Augmentation')
parser.add_argument('--bblur', action='store_true', default=False,
                    help='Use Bilateral Blur Augmentation')
parser.add_argument('--lr_schedule', type=str, default='poly',
                    help='name of lr schedule: poly')
parser.add_argument('--poly_exp', type=float, default=0.9,
                    help='polynomial LR exponent')
parser.add_argument('--bs_mult', type=int, default=1,
                    help='Batch size for training per gpu')
parser.add_argument('--bs_mult_val', type=int, default=1,
                    help='Batch size for Validation per gpu')
parser.add_argument('--crop_size', type=int, default=720,
                    help='training crop size')
parser.add_argument('--pre_size', type=int, default=None,
                    help='resize image shorter edge to this before augmentation')
parser.add_argument('--scale_min', type=float, default=0.5,
                    help='dynamically scale training images down to this size')
parser.add_argument('--scale_max', type=float, default=2.0,
                    help='dynamically scale training images up to this size')
parser.add_argument('--weight_decay', type=float, default=5e-4)
parser.add_argument('--momentum', type=float, default=0.9)
parser.add_argument('--snapshot', type=str, default=None)
parser.add_argument('--restore_optimizer', action='store_true', default=False)

parser.add_argument('--city_mode', type=str, default='train',
                    help='experiment directory date name')
parser.add_argument('--date', type=str, default='default',
                    help='experiment directory date name')
parser.add_argument('--exp', type=str, default='default',
                    help='experiment directory name')
parser.add_argument('--tb_tag', type=str, default='',
                    help='add tag to tb dir')
parser.add_argument('--ckpt', type=str, default='logs/ckpt',
                    help='Save Checkpoint Point')
parser.add_argument('--tb_path', type=str, default='logs/tb',
                    help='Save Tensorboard Path')
parser.add_argument('--syncbn', action='store_true', default=False,
                    help='Use Synchronized BN')
parser.add_argument('--dump_augmentation_images', action='store_true', default=False,
                    help='Dump Augmentated Images for sanity check')
parser.add_argument('--test_mode', action='store_true', default=False,
                    help='Minimum testing to verify nothing failed, ' +
                    'Runs code for 1 epoch of train and val')
parser.add_argument('-wb', '--wt_bound', type=float, default=1.0,
                    help='Weight Scaling for the losses')
parser.add_argument('--maxSkip', type=int, default=0,
                    help='Skip x number of  frames of video augmented dataset')
parser.add_argument('--scf', action='store_true', default=False,
                    help='scale correction factor')
parser.add_argument('--dist_url', default='tcp://127.0.0.1:', type=str,
                    help='url used to set up distributed training')

parser.add_argument('--wt_layer', nargs='*', type=int, default=[0,0,0,0,0,0,0],
                    help='0: None, 1: IW/IRW, 2: ISW, 3: IS, 4: IN (IBNNet: 0 0 4 4 4 0 0)')
parser.add_argument('--wt_reg_weight', type=float, default=0.0)
parser.add_argument('--relax_denom', type=float, default=2.0)
parser.add_argument('--clusters', type=int, default=50)
parser.add_argument('--trials', type=int, default=10)
parser.add_argument('--dynamic', action='store_true', default=False)

parser.add_argument('--image_in', action='store_true', default=False,
                    help='Input Image Instance Norm')
parser.add_argument('--cov_stat_epoch', type=int, default=5,
                    help='cov_stat_epoch')
parser.add_argument('--visualize_feature', action='store_true', default=False,
                    help='Visualize intermediate feature')
parser.add_argument('--use_wtloss', action='store_true', default=False,
                    help='Automatic setting from wt_layer')
parser.add_argument('--use_isw', action='store_true', default=False,
                    help='Automatic setting from wt_layer')

args_base, unknown = parser.parse_known_args()
args_isw, unknown = parser.parse_known_args()


# Enable CUDNN Benchmarking optimization
#torch.backends.cudnn.benchmark = True
random_seed = cfg.RANDOM_SEED  #304
torch.manual_seed(random_seed)
torch.cuda.manual_seed(random_seed)
torch.cuda.manual_seed_all(random_seed) # if use multi-GPU
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False
torch.set_num_threads(8)
np.random.seed(random_seed)
random.seed(random_seed)

args_base.world_size = 1

# Test Mode run two epochs with a few iterations of training and val
if args_base.test_mode:
    args_base.max_epoch = 2

if 'WORLD_SIZE' in os.environ:
    # args.apex = int(os.environ['WORLD_SIZE']) > 1
    args_base.world_size = int(os.environ['WORLD_SIZE'])
    print("Total world size: ", int(os.environ['WORLD_SIZE']))

args_base.dist_url = args_base.dist_url + str(8000 + (int(time.time()%1000))//10)

def load_models(BASELINE_PATH, ISW_PATH):
    """
    Main Function
    """
    # Set up the Arguments, Tensorboard Writer, Dataloader, Loss Fn, Optimizer
    assert_and_infer_cfg(args_base)
    #prep_experiment(args, parser)
    writer = None
    args_base.wt_layer = [0, 0, 0, 0, 0, 0, 0]
    args_isw.wt_layer = [0, 0, 2, 2, 2, 0, 0]
    
    args_base.snapshot = os.path.join(BASELINE_PATH)
    args_isw.snapshot = os.path.join(ISW_PATH)

    criterion, criterion_val = loss.get_loss(args_base)
    criterion_aux = loss.get_loss_aux(args_base)

    net_base = network.get_net(args_base, criterion, criterion_aux)
    net_isw = network.get_net(args_isw, criterion, criterion_aux)

    loaded_dict_base = torch.load(args_base.snapshot, map_location=torch.device('cpu'))
    loaded_dict_isw = torch.load(args_isw.snapshot, map_location=torch.device('cpu'))

    net_base = optimizer.forgiving_state_restore(net_base, loaded_dict_base)
    net_isw = optimizer.forgiving_state_restore(net_isw, loaded_dict_isw)

    net_base.eval()
    net_isw.eval()
    
    return net_base, net_isw

def load_models(SRC_DIR, arch="r50", trained_on="cityscapes"):
    """
    Main Function
    """
    # Set up the Arguments, Tensorboard Writer, Dataloader, Loss Fn, Optimizer
    assert_and_infer_cfg(args_base)
    #prep_experiment(args, parser)
    writer = None
    args_base.wt_layer = [0, 0, 0, 0, 0, 0, 0]
    args_isw.wt_layer = [0, 0, 2, 2, 2, 0, 0]

    if trained_on != "cityscapes" and trained_on != "gtav":
        raise ValueError("Invalid trained_on dataset")
    if arch != "r50" and arch != "r101":
        raise ValueError("Invalid architecture! Choose between r50 or r101")

    if arch == "r101":
        args_isw.arch = 'network.deepv3.DeepR101V3PlusD_OS8'
        args_base.arch = 'network.deepv3.DeepR101V3PlusD_OS8'
    
    args_base.snapshot = os.path.join(SRC_DIR, r"base_" + trained_on + "_" + arch + "_state_dict.pth")
    args_isw.snapshot = os.path.join(SRC_DIR, r"isw_" + trained_on + "_" + arch + "_state_dict.pth")

    criterion, criterion_val = loss.get_loss(args_base)
    criterion_aux = loss.get_loss_aux(args_base)

    net_base = network.get_net(args_base, criterion, criterion_aux)
    net_isw = network.get_net(args_isw, criterion, criterion_aux)

    loaded_dict_base = torch.load(args_base.snapshot, map_location=torch.device('cpu'))
    loaded_dict_isw = torch.load(args_isw.snapshot, map_location=torch.device('cpu'))

    net_base = optimizer.forgiving_state_restore(net_base, loaded_dict_base)
    net_isw = optimizer.forgiving_state_restore(net_isw, loaded_dict_isw)

    net_base.eval()
    net_isw.eval()
    
    return net_base, net_isw

def predict(model, input):

    trans = transforms.Compose([transforms.ToTensor()])
    input = trans(input)

    if len(input.shape) < 4:
        input = input.unsqueeze(0)

    # Extract predictions
    outputs = model(input)
    predictions = outputs.data.max(1)[1]

    predictions = predictions.squeeze(0).detach().numpy()
    segmap = label2rgb(predictions)
    return segmap


def label2rgb(pred_np):

    r = pred_np.copy()
    g = pred_np.copy()
    b = pred_np.copy()
    for l in range(0, 19):
        r[pred_np == l] = class_colors[l][0]
        g[pred_np == l] = class_colors[l][1]
        b[pred_np == l] = class_colors[l][2]

    rgb = np.zeros((pred_np.shape[0], pred_np.shape[1], 3))
    rgb[:, :, 0] = r
    rgb[:, :, 1] = g
    rgb[:, :, 2] = b
    return rgb.astype(np.uint8)



